'use strict';

const AWS = require('aws-sdk');

// BEGIN : To be removed from lambda deployment
AWS.config.update({region:'us-east-1'});
const msgPut = {
    "body-json": {
        "data": {
            "type": "photos",
            "attributes": {
                "title": "Ember Hamster",
                "src": "http://example.com/images/productivity.png"
            },
            "relationships": {
                "photographer": {
                    "data": { "type": "people", "id": "9" }
                }
            }
        }
    },
    "params": {
        "path": {
            "type": "photos"
        },
        "querystring": {},
        "header": {
            "Accept": "*/*",
            "accept-encoding": "gzip, deflate",
            "cache-control": "no-cache",
            "Content-Type": "application/json",
            "Host": "9o6yk638sk.execute-api.us-east-1.amazonaws.com",
            "Postman-Token": "ca9ccefd-42e1-4bad-85bd-620050ef1770",
            "User-Agent": "PostmanRuntime/7.4.0",
            "X-Amz-Date": "20181122T091255Z",
            "X-Amzn-Trace-Id": "Root=1-5bf67317-3e732e202ed67512573c7ca4",
            "X-Forwarded-For": "195.5.224.170",
            "X-Forwarded-Port": "443",
            "X-Forwarded-Proto": "https"
        }
    },
    "stage-variables": {},
    "context": {
        "account-id": "",
        "api-id": "9o6yk638sk",
        "api-key": "",
        "authorizer-principal-id": "",
        "caller": "",
        "cognito-authentication-provider": "",
        "cognito-authentication-type": "",
        "cognito-identity-id": "",
        "cognito-identity-pool-id": "",
        "httpMethod": "POST",
        "stage": "default",
        "source-ip": "195.5.224.170",
        "user": "",
        "user-agent": "PostmanRuntime/7.4.0",
        "user-arn": "",
        "request-id": "cc5001d5-ee36-11e8-8ae7-273051a62acd",
        "resource-id": "anwfak",
        "resource-path": "/EmberDataServerless/{type}"
    }
};
// END

const dynamo = new AWS.DynamoDB.DocumentClient();
const tableName = 'EmberDataServerlessTable';
const EPOCH = 1300000000000;

// Instagram inspired --> https://instagram-engineering.com/sharding-ids-at-instagram-1cf5a71e5a5c
function generateRowId(subid) {
  var ts = new Date().getTime() - EPOCH; // limit to recent
  // 41 bits for time in milliseconds (gives us 41 years of IDs with a custom epoch
  var randid = Math.floor(Math.random() * 512);
  ts = (ts * 64);   // bit-shift << 6
  // Given shard (if any...)
  ts = ts + subid;
  // random value
  return (ts * 512) + (randid % 512);
}

const createObject = (obj) => {
    let objout = {  "type": obj.ObjectType,
                    "id": obj.Id,
                    "attributes": {},
                    "relationships": {}
                };
    for (var attr in obj) {
        if (attr !== 'ObjectType' && attr !== 'Id') {
            if(!attr.endsWith('_id')) {
                objout.attributes[attr] = obj[attr]; 
            } else {
                let relationDetails = attr.split('_');
                let relationId = Number(obj[attr]);
                objout.relationships[relationDetails[0]] = {
                        "links": {
                            "self": "/"+obj.ObjectType+"/"+obj.Id+"/relationships/"+relationDetails[0],
                            "related": "/"+obj.ObjectType+"/"+obj.Id+"/"+relationDetails[0]
                        },
                        "data": {"type": relationDetails[1], "id": relationId}
                    };
                // TODO what's about if relation is not a belongsTo but a hasMany...
            }         
        }
    }
    return objout;
}

const createData = (data) => {
    if (Array.isArray(data)) {
        let outdata = [];
        for (let i=0;i<data.length;i++){
            outdata.push(createObject(data[i]));
         }
         return outdata;
    } else {
        return createObject(data);
    }
}

const createRelationships = (data) => {
    return {};
}

const createResponse = (statusCode, body) => {
    return {
        'statusCode': statusCode,
        'data': createData(body),
        'relationships': createRelationships(body)
    }
};

const getMethod = (event, context, callback) => {

    let params = {
        TableName: tableName,
    },
    type = event.params.path.type,
    id = Number(event.params.path.id),
    dbGet = {};

    if (id) {
        params.Key = {
            'ObjectType': type,
            'Id': id
        };
        dbGet = (params) => { return dynamo.get(params).promise() };
        console.log('EmberDataServerless lambda GET single value with params: ', params);

    } else {
        params.KeyConditionExpression = 'ObjectType = :objectType';
        params.ExpressionAttributeValues = { ':objectType': type };
        dbGet = (params) => { return dynamo.query(params).promise() };
        console.log('EmberDataServerless lambda GET multiple values with params: ', params);
    }

    dbGet(params).then( (data) => {
        console.log('EmberDataServerless lambda GET data received: ', data);

        if (id && !data.Item) {
            callback(null, createResponse(404, "ITEM NOT FOUND"));
            return;
        } else if (id && data.Item) {
            console.log(`RETRIEVED ITEM SUCCESSFULLY WITH doc = ${data.Item}`);
            callback(null, createResponse(200, data.Item));
        } else {
            console.log('SCANNING TABLE');
            console.log(`RETRIEVED ITEMS SUCCESSFULLY WITH doc = ${data.Items}`);
            callback(null, createResponse(200, data.Items));
        }
        
    }).catch( (err) => { 
        console.log(`GET ITEM FAILED FOR Entry = ${params}, WITH ERROR: ${err}`);
        callback(null, createResponse(500, err));
    });
};

const putMethod = (event, context, callback) => {
    const body=event['body-json'];
    const attrs = body.data.attributes;
    const relations = body.data.relationships;
    // Without any body content, there is nothing to put...
    if (!body || !attrs) {
        callback(null, createResponse(500, 'No content found in body'));
        return;
    }
    // Retrieving the type and generating a new id for created item
    let type = event.params.path.type,
        id = generateRowId(1);
    // Final content contains at least these two fields + the atributes
    let content = {
        "ObjectType": type,
        "Id": id
    };
    // Adding attributes as column in dynamoDb
    for (var prop in attrs) {
        content[prop] = attrs[prop];
    }
    // Dealing with relationships if any
    if (relations){
        for (var relName in relations) {
            let relData = relations[relName]["data"];
            let newCol;
            if (!Array.isArray(relData)) {
                newCol = relName+'_'+relData["type"]+'_id';
                content[newCol] = relData["id"];
            } else {
                for (var i=0; i<relData.length;i++){
                    let currentData = relData[i];
                    newCol = relName+'_'+currentData["type"]+'_id';
                    content[newCol] = currentData["id"];
                }
            }
        }
    }
    
    const entry = {
        TableName: 'EmberDataServerlessTable',
        Item: content
    };
    console.log('Try saving entity of type '+type+' and content '+JSON.stringify(entry));
    //let dbPut = (entry) => { return dynamo.put(entry).promise() };
    dynamo.put(entry, function(err, data) {
        if (err) {
            console.log("Error", err);
            callback(null, createResponse(500, 'Error '+err));
        } else {
            body.data.id = id;
            body['statusCode'] = 200;
            console.log(`PUT ITEM SUCCEEDED WITH data=`+JSON.stringify(body));
            callback(null, body);
        }
    });
};

const deleteMethod = (event, context, callback) => {
    let type = event.params.path.type,
        id = Number(event.params.path.id),
        params = {
            'TableName': tableName,
            'Key': {
                'ObjectType': type,
                'Id': id
            },
            'ReturnValues': 'ALL_OLD'
        };
    let dbDelete = (params) => { return dynamo.delete(params).promise() };
    dbDelete(params).then( (data) => {
        if (!data.Attributes) {
            callback(null, createResponse(404, "ITEM NOT FOUND FOR DELETION"));
            return;
        }
        console.log(`DELETED ITEM OF TYPE ${type} SUCCESSFULLY WITH id = ${id}`);
        callback(null, createResponse(200,data));
    }).catch( (err) => { 
        console.log(`DELETE ITEM OF TYPE ${type} FAILED FOR id = ${id}, WITH ERROR: ${err}`);
        callback(null, createResponse(500, err));
    });
};


const callback = (evt, msg) => {console.log(msg);};

//putMethod(msgPut,{},callback);

const msgGet = {
    "params": {
        "path": {
            "type": "photos"
        },
        "querystring": {},
        "header": {}
    },
    "context": {
        "httpMethod": "GET",
        "account-id": "995466093880",
        "api-id": "9o6yk638sk",
        "api-key": "test-invoke-api-key",
        "authorizer-principal-id": "",
        "caller": "995466093880",
        "cognito-authentication-provider": "",
        "cognito-authentication-type": "",
        "cognito-identity-id": "",
        "cognito-identity-pool-id": "",
        "stage": "test-invoke-stage",
        "source-ip": "test-invoke-source-ip",
        "user": "995466093880",
        "user-agent": "aws-internal/3 aws-sdk-java/1.11.447 Linux/4.9.124-0.1.ac.198.71.329.metal1.x86_64 OpenJDK_64-Bit_Server_VM/25.192-b12 java/1.8.0_192",
        "user-arn": "arn:aws:iam::995466093880:root",
        "request-id": "b2b534a0-ef42-11e8-bbf2-b564234403bf",
        "resource-id": "anwfak",
        "resource-path": "/EmberDataServerless/{type}"
    }
};

getMethod(msgGet,{},callback)